import gst
import os

PWD = os.path.dirname(os.path.abspath(__file__ + '/..'))
__PLAYER__ = gst.element_factory_make("playbin", "player")

def play(sound, path=None):
    #global PWD
    global __PLAYER__
    __PLAYER__.set_state(gst.STATE_NULL)
    __PLAYER__ = gst.element_factory_make("playbin", "player")

    if path:
        __PLAYER__.set_property('uri', 'file://%s/%s.wav' % (path, sound))
    else:
        __PLAYER__.set_property('uri', 'file://%s/sounds/%s.wav' % (PWD, sound))

    __PLAYER__.set_state(gst.STATE_PLAYING)


from twisted.python import log
import signal
gobjectnotimported = False


try:
    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default=True)
except:
    gobjectnotimported = True
    pass


import gobject
import pygst
pygst.require('0.10')
gobject.threads_init()

from lisa.client import lib

from twisted.internet import ssl, utils
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.internet.defer import inlineCallbacks, DeferredQueue
from twisted.protocols.basic import LineReceiver
from twisted.application import internet, service
import json, os
from OpenSSL import SSL
import platform
from twisted.application.internet import TimerService

import pkg_resources
configuration = json.load(open(pkg_resources.resource_filename(__name__, 'configuration/lisa.json.sample')))
path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)
soundfile = '/tmp/lisa-output.wav'
sound_queue = DeferredQueue()
botname = ""


@inlineCallbacks
def SoundWorker():
    data = yield sound_queue.get()
    command_create = ('-w', soundfile,
               '-l', configuration['lang'], '"'+ data.encode('UTF-8') + '"')
    create_sound = yield utils.getProcessOutputAndValue('/usr/bin/pico2wave', path='/usr/bin', args=command_create)
    if os.path.isfile(soundfile):
        play_sound = yield lib.player.play('lisa-output', path='/tmp')
        os.remove(soundfile)
    else:
        log.err("There was an error creating the output file %s" % soundfile)


class LisaClient(LineReceiver):
    def __init__(self,factory):
        global gobjectnotimported
        self.gobjectnotimported = gobjectnotimported
        self.factory = factory
        self.bot_name = "lisa"
        botname = "lisa"
        self.listener = None

    def sendMessage(self, message, type='chat'):
        if configuration['debug']['debug_output']:
            log.msg('OUTPUT: "from": ' + unicode(platform.node()) + ',"type": ' + type + ', "body": ' + unicode(message) +
                    ', "zone": ' + configuration['zone']
            )
        self.sendLine(json.dumps(
            {"from": unicode(platform.node()), "type": type, "body": unicode(message), "zone": configuration['zone']})
        )

    # TODO clean the global botname. I needed to use it, but I think we can use the self.bot_name instead
    def lineReceived(self, data):
        datajson = json.loads(data)
        if configuration['debug']['debug_input']:
            log.msg("INPUT: " + unicode(datajson))
        if 'type' in datajson:
            if datajson['type'] == 'chat':
                sound_queue.put(datajson['body'])
            elif datajson['type'] == 'command':
                if datajson['command'] == 'LOGIN':
                    log.msg("I found login")
                    self.bot_name = unicode(datajson['bot_name'])
                    global botname
                    botname = unicode(datajson['bot_name'])
                    log.msg("setting botname to %s" % self.bot_name)
                    sound_queue.put(datajson['body'])
                    if not 'nolistener' in datajson and not gobjectnotimported:
                        self.listener = lib.Listener(lisaclient=self, botname=botname)
                # TODO seems a bit more complicated than I thought. I think the reply will be another type like "answer"
                # TODO and will contains a unique ID. On server side, the question will be stored in mongodb so it will
                # TODO let possible the multi user / multi client. Questions need to implement a lifetime too.
                # TODO For the soundqueue, I will need a callback system to be sure to play the audio before recording
                elif datajson['command'] == 'ASK':
                    sound_queue.put(datajson['body'])
                    if not 'nolistener' in datajson and not gobjectnotimported:
                        self.listener.record()

        else:
            sound_queue.put(datajson['body'])

    def connectionMade(self):
        log.msg('Connected to Lisa.')
        if configuration['enable_secure_mode']:
            ctx = ClientTLSContext()
            self.transport.startTLS(ctx, self.factory)
        self.sendMessage(message='LOGIN', type='command')
        if not gobjectnotimported:
            #init gobject threads
            gobject.threads_init()
            #we want a main loop
            main_loop = gobject.MainLoop()
            #handle sigint
            signal.signal(signal.SIGINT, signal.SIG_DFL)




class LisaClientFactory(ReconnectingClientFactory):
    def startedConnecting(self, connector):
        log.msg('Started to connect.')

    def buildProtocol(self, addr):
        self.protocol = LisaClient(self)
        log.msg('Resetting reconnection delay')
        self.resetDelay()
        return self.protocol

    def clientConnectionLost(self, connector, reason):
        log.err('Lost connection.  Reason:', reason)
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)

    def clientConnectionFailed(self, connector, reason):
        log.err('Connection failed. Reason:', reason)
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)


class ClientTLSContext(ssl.ClientContextFactory):
    isClient = 1
    def getContext(self):
        return SSL.Context(SSL.TLSv1_METHOD)


class CtxFactory(ssl.ClientContextFactory):
    def getContext(self):
        self.method = SSL.SSLv23_METHOD
        ctx = ssl.ClientContextFactory.getContext(self)
        ctx.use_certificate_file(os.path.normpath(dir_path + '/' + 'configuration/ssl/client.crt'))
        ctx.use_privatekey_file(os.path.normpath(dir_path + '/' + 'configuration/ssl//client.key'))

        return ctx

# Creating MultiService
application = service.Application("LISA-Client")

def makeService(config):
    global configuration

    if 'configuration' in config.keys():
        configuration = json.load(open(config['configuration']))

    multi = service.MultiService()
    multi.setServiceParent(application)

    sound_service = TimerService(0.01, SoundWorker)
    sound_service.setServiceParent(multi)

    LisaFactory = LisaClientFactory()

    if configuration['enable_secure_mode']:
        lisaclientService =  internet.TCPClient(configuration['lisa_url'], configuration['lisa_engine_port_ssl'], LisaFactory, CtxFactory())
    else:
        lisaclientService =  internet.TCPClient(configuration['lisa_url'], configuration['lisa_engine_port'], LisaFactory)

    lisaclientService.setServiceParent(multi)
    return multi
#makeService(config=[])


from lisa.client.service import LisaClientFactory
from twisted.trial import unittest
from twisted.test import proto_helpers

import json
import platform


class LisaClientTestCase(unittest.TestCase):
    def setUp(self):
        factory = LisaClientFactory()
        self.proto = factory.buildProtocol(('127.0.0.1', 0))
        self.tr = proto_helpers.StringTransport()
        self.proto.makeConnection(self.tr)

    def convertjson(self, type=None, message=None, command=None, argdict=None):
        jsondict = {"from": "LISA Server"}
        if message:
            jsondict['body'] = unicode(message)
        if type:
            jsondict['type'] = type
        if command:
            jsondict['command'] = command
        if argdict:
            for k,v in argdict.items():
                jsondict[k] = v

        return json.dumps(jsondict)

    #def test_receivechat(self):
    #    json = self.convertjson(message="")


    def test_receive_command_login(self):
        message = "The client " + unicode(platform.node()) + " joined the zone"
        argdict = {"bot_name": "test", "nolistener": True}
        json = self.convertjson(message=message, type="command", command="LOGIN", argdict=argdict)
        self.proto.lineReceived(data=json)
        self.assertEqual(self.proto.bot_name, "test")



# It fixes a bug with the pocketpshinx import. The first time it fails, but the second import is ok.
try:
    import pocketsphinx
except:
    pass
import pocketsphinx
from wit import Wit
import gobject
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)
from twisted.python import log
import json, os
import pygst
pygst.require('0.10')
gobject.threads_init()
import gst
from lisa.client.lib import player
from lisa.client.lib.recorder import Recorder


path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)
if os.path.exists('/etc/lisa/client/configuration/lisa.json'):
    configuration = json.load(open('/etc/lisa/client/configuration/lisa.json'))
else:
    configuration = json.load(open(os.path.normpath(dir_path + '/../' + 'configuration/lisa.json')))


class Listener:
    def __init__(self, lisaclient, botname):

        self.configuration = configuration
        self.recording_state = False
        self.botname = botname
        self.lisaclient = lisaclient
        self.failed = 0
        self.keyword_identified = 0
        self.recorder = Recorder(configuration=self.configuration)
        self.wit = Wit(self.configuration['wit_token'])

        # The goal is to listen for a keyword. When I have this keyword, it records for few seconds and stream the
        # flow to Wit which answer by json.

        self.pipeline = gst.parse_launch('alsasrc ! audioconvert ! audioresample '
				+ '! vader name=vad auto-threshold=true '
				+ '! pocketsphinx name=asr ! fakesink')

        self.vader = self.pipeline.get_by_name('vad')

        asr = self.pipeline.get_by_name('asr')
        if os.path.isfile('/var/lib/lisa/client/pocketsphinx/lisa.dic'):
            asr.set_property("dict", '/var/lib/lisa/client/pocketsphinx/lisa.dic')
            asr.set_property("lm", '/var/lib/lisa/client/pocketsphinx/lisa.lm')
        else:
            asr.set_property("dict", "%s/lib/pocketsphinx/lisa.dic" % os.path.normpath(dir_path + '/../'))
            asr.set_property("lm", "%s/lib/pocketsphinx/lisa.lm" % os.path.normpath(dir_path + '/../'))
        asr.connect('result', self.result)
        asr.set_property('configured', True)

        self.ps = pocketsphinx.Decoder(boxed=asr.get_property('decoder'))

        self.pipeline.set_state(gst.STATE_PLAYING)

    def result(self, asr, hyp, uttid):
        print hyp.lower()
        if hyp.lower() == self.botname.lower() and not self.recording_state:
            struct = gst.Structure('result')
            dec_text, dec_uttid, dec_score = self.ps.get_hyp()

            if dec_score >= self.configuration['keyword_score']:
                log.msg("======================")
                log.msg("%s keyword detected" % self.botname)
                log.msg("score: {}".format(dec_score))

                self.failed = 0
                self.keyword_identified = 1
                self.record()
            else:
                log.msg("I recognized the %s keyword but I think it's a false positive according the %s score" %
                        (self.botname.lower(), dec_score))

    def cancel_listening(self):
        log.msg("cancel_listening : player.play('pi-cancel')")
        player.play('pi-cancel')
        self.recording_state = False

    # sound recording
    def record(self):
        self.pipeline.set_state(gst.STATE_PAUSED)
        self.recording_state = True
        # This content type (raw) allow to send data from mic directly to Wit and stream chunks
        # thanks to the generator
        CONTENT_TYPE = 'raw;encoding=signed-integer;bits=16;rate=16000;endian=little'
        result = self.wit.post_speech(data=self.recorder.capture_audio(), content_type=CONTENT_TYPE)
        player.play('pi-cancel')
        log.msg(" * Contacting Google")
        if len(result) == 0:
            log.msg("cancel_listening : player.play('pi-cancel')")
            player.play('pi-cancel')
        else:
            log.msg(result)
            self.lisaclient.sendMessage(result['msg_body'])
        self.recording_state = False
        self.pipeline.set_state(gst.STATE_PLAYING)

    def get_pipeline(self):
        return self.pipeline


import alsaaudio
from lisa.client.lib import player

class Loader():
    def __init__(self, end, every):
        self.end = end / every - 1
        self.every = every
        self.progress = '.'

    def _clear_line(self):
        ''' Clear one terminal line '''
        print '\x1b[1A',

    def __call__(self, i):
        ''' Print loader for the ith time '''
        if not i % self.every:
            return
        self._clear_line()
        print '[Recording]' + self.progress * (i/2) + ' ' * (self.end - (i/2)) + '[Recording]'


class Recorder:
    def __init__(self, configuration):
        self.configuration = configuration

        # Microphone stream config.
        self.chunk = 1024  # CHUNKS of bytes to read each time from mic
        self.channels = 1
        self.rate = 16000

        self.RECORD_SECONDS = configuration['record_seconds']
        self.SLEEP_TIME = configuration['sleep_time']


    def capture_audio(self):
        def setup_mic():
            inp = alsaaudio.PCM(alsaaudio.PCM_CAPTURE, alsaaudio.PCM_NORMAL)
            inp.setchannels(self.channels)
            inp.setrate(self.rate)
            inp.setformat(alsaaudio.PCM_FORMAT_S16_LE)
            inp.setperiodsize(self.chunk)
            return inp
        player.play('pi-listening')
        loader = Loader(self.SLEEP_TIME * self.RECORD_SECONDS, 2)
        inp = setup_mic()

        print "\n[*]> Starting Recording\n"
        for i in xrange(0, self.rate / self.chunk * self.RECORD_SECONDS):
            loader(i)
            _, data = inp.read()
            yield data
        print "[*]> Ready Recognize Voice\n"
